from collections import defaultdict
from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray
from tqdm import tqdm
from valor_lite.classification.annotation import Classification
from valor_lite.classification.computation import (
    compute_confusion_matrix,
    compute_precision_recall_rocauc,
)
from valor_lite.classification.metric import Metric, MetricType
from valor_lite.classification.utilities import (
    unpack_confusion_matrix_into_metric_list,
    unpack_precision_recall_rocauc_into_metric_lists,
)

"""
Usage
-----

manager = DataLoader()
manager.add_data(
    groundtruths=groundtruths,
    predictions=predictions,
)
evaluator = manager.finalize()

metrics = evaluator.evaluate()

f1_metrics = metrics[MetricType.F1]
accuracy_metrics = metrics[MetricType.Accuracy]

filter_mask = evaluator.create_filter(datum_uids=["uid1", "uid2"])
filtered_metrics = evaluator.evaluate(filter_mask=filter_mask)
"""


@dataclass
class Filter:
    indices: NDArray[np.int32]
    label_metadata: NDArray[np.int32]
    n_datums: int


class Evaluator:
    """
    Classification Evaluator
    """

    def __init__(self):

        # metadata
        self.n_datums = 0
        self.n_groundtruths = 0
        self.n_predictions = 0
        self.n_labels = 0

        # datum reference
        self.uid_to_index: dict[str, int] = dict()
        self.index_to_uid: dict[int, str] = dict()

        # label reference
        self.label_to_index: dict[str, int] = dict()
        self.index_to_label: dict[int, str] = dict()

        # computation caches
        self._detailed_pairs = np.array([])
        self._label_metadata = np.array([], dtype=np.int32)
        self._label_metadata_per_datum = np.array([], dtype=np.int32)

    @property
    def ignored_prediction_labels(self) -> list[str]:
        """
        Prediction labels that are not present in the ground truth set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (plabels - glabels)
        ]

    @property
    def missing_prediction_labels(self) -> list[str]:
        """
        Ground truth labels that are not present in the prediction set.
        """
        glabels = set(np.where(self._label_metadata[:, 0] > 0)[0])
        plabels = set(np.where(self._label_metadata[:, 1] > 0)[0])
        return [
            self.index_to_label[label_id] for label_id in (glabels - plabels)
        ]

    @property
    def metadata(self) -> dict:
        """
        Evaluation metadata.
        """
        return {
            "n_datums": self.n_datums,
            "n_groundtruths": self.n_groundtruths,
            "n_predictions": self.n_predictions,
            "n_labels": self.n_labels,
            "ignored_prediction_labels": self.ignored_prediction_labels,
            "missing_prediction_labels": self.missing_prediction_labels,
        }

    def create_filter(
        self,
        datum_uids: list[str] | NDArray[np.int32] | None = None,
        labels: list[str] | NDArray[np.int32] | None = None,
    ) -> Filter:
        """
        Creates a boolean mask that can be passed to an evaluation.

        Parameters
        ----------
        datum_uids : list[str] | NDArray[np.int32], optional
            An optional list of string uids or a numpy array of uid indices.
        labels : list[str] | NDArray[np.int32], optional
            An optional list of labels or a numpy array of label indices.

        Returns
        -------
        Filter
            A filter object that can be passed to the `evaluate` method.
        """
        n_rows = self._detailed_pairs.shape[0]

        n_datums = self._label_metadata_per_datum.shape[1]
        n_labels = self._label_metadata_per_datum.shape[2]

        mask_pairs = np.ones((n_rows, 1), dtype=np.bool_)
        mask_datums = np.ones(n_datums, dtype=np.bool_)
        mask_labels = np.ones(n_labels, dtype=np.bool_)

        if datum_uids is not None:
            if isinstance(datum_uids, list):
                datum_uids = np.array(
                    [self.uid_to_index[uid] for uid in datum_uids],
                    dtype=np.int32,
                )
            mask = np.zeros_like(mask_pairs, dtype=np.bool_)
            mask[
                np.isin(self._detailed_pairs[:, 0].astype(int), datum_uids)
            ] = True
            mask_pairs &= mask

            mask = np.zeros_like(mask_datums, dtype=np.bool_)
            mask[datum_uids] = True
            mask_datums &= mask

        if labels is not None:
            if isinstance(labels, list):
                labels = np.array(
                    [self.label_to_index[label] for label in labels]
                )
            mask = np.zeros_like(mask_pairs, dtype=np.bool_)
            mask[
                np.isin(self._detailed_pairs[:, 1].astype(int), labels)
            ] = True
            mask_pairs &= mask

            mask = np.zeros_like(mask_labels, dtype=np.bool_)
            mask[labels] = True
            mask_labels &= mask

        mask = mask_datums[:, np.newaxis] & mask_labels[np.newaxis, :]
        label_metadata_per_datum = self._label_metadata_per_datum.copy()
        label_metadata_per_datum[:, ~mask] = 0

        label_metadata = np.zeros_like(self._label_metadata, dtype=np.int32)
        label_metadata = np.transpose(
            np.sum(
                label_metadata_per_datum,
                axis=1,
            )
        )

        n_datums = int(np.sum(label_metadata[:, 0]))

        return Filter(
            indices=np.where(mask_pairs)[0],
            label_metadata=label_metadata,
            n_datums=n_datums,
        )

    def compute_precision_recall_rocauc(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        filter_: Filter | None = None,
    ) -> dict[MetricType, list]:
        """
        Performs an evaluation and returns metrics.

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        dict[MetricType, list]
            A dictionary mapping MetricType enumerations to lists of computed metrics.
        """

        # apply filters
        data = self._detailed_pairs
        label_metadata = self._label_metadata
        n_datums = self.n_datums
        if filter_ is not None:
            data = data[filter_.indices]
            label_metadata = filter_.label_metadata
            n_datums = filter_.n_datums

        results = compute_precision_recall_rocauc(
            data=data,
            label_metadata=label_metadata,
            score_thresholds=np.array(score_thresholds),
            hardmax=hardmax,
            n_datums=n_datums,
        )

        return unpack_precision_recall_rocauc_into_metric_lists(
            results=results,
            score_thresholds=score_thresholds,
            hardmax=hardmax,
            label_metadata=label_metadata,
            index_to_label=self.index_to_label,
        )

    def compute_confusion_matrix(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        number_of_examples: int = 0,
        filter_: Filter | None = None,
    ) -> list[Metric]:
        """
        Computes a detailed confusion matrix..

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        number_of_examples : int, default=0
            The number of examples to return per count.
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        list[Metric]
            A list of confusion matrices.
        """

        # apply filters
        data = self._detailed_pairs
        label_metadata = self._label_metadata
        if filter_ is not None:
            data = data[filter_.indices]
            label_metadata = filter_.label_metadata

        if data.size == 0:
            return list()

        results = compute_confusion_matrix(
            data=data,
            label_metadata=label_metadata,
            score_thresholds=np.array(score_thresholds),
            hardmax=hardmax,
            n_examples=number_of_examples,
        )

        return unpack_confusion_matrix_into_metric_list(
            results=results,
            score_thresholds=score_thresholds,
            number_of_examples=number_of_examples,
            index_to_uid=self.index_to_uid,
            index_to_label=self.index_to_label,
        )

    def evaluate(
        self,
        score_thresholds: list[float] = [0.0],
        hardmax: bool = True,
        number_of_examples: int = 0,
        filter_: Filter | None = None,
    ) -> dict[MetricType, list[Metric]]:
        """
        Computes a detailed confusion matrix..

        Parameters
        ----------
        score_thresholds : list[float]
            A list of score thresholds to compute metrics over.
        hardmax : bool
            Toggles whether a hardmax is applied to predictions.
        number_of_examples : int, default=0
            The number of examples to return per count.
        filter_ : Filter, optional
            An optional filter object.

        Returns
        -------
        dict[MetricType, list[Metric]]
            Lists of metrics organized by metric type.
        """

        metrics = self.compute_precision_recall_rocauc(
            score_thresholds=score_thresholds,
            hardmax=hardmax,
            filter_=filter_,
        )

        metrics[MetricType.ConfusionMatrix] = self.compute_confusion_matrix(
            score_thresholds=score_thresholds,
            hardmax=hardmax,
            number_of_examples=number_of_examples,
            filter_=filter_,
        )

        return metrics


class DataLoader:
    """
    Classification DataLoader.
    """

    def __init__(self):
        self._evaluator = Evaluator()
        self.groundtruth_count = defaultdict(lambda: defaultdict(int))
        self.prediction_count = defaultdict(lambda: defaultdict(int))

    def _add_datum(self, uid: str) -> int:
        """
        Helper function for adding a datum to the cache.

        Parameters
        ----------
        uid : str
            The datum uid.

        Returns
        -------
        int
            The datum index.
        """
        if uid not in self._evaluator.uid_to_index:
            index = len(self._evaluator.uid_to_index)
            self._evaluator.uid_to_index[uid] = index
            self._evaluator.index_to_uid[index] = uid
        return self._evaluator.uid_to_index[uid]

    def _add_label(self, label: str) -> int:
        """
        Helper function for adding a label to the cache.

        Parameters
        ----------
        label : str
            A string representing a label.

        Returns
        -------
        int
            Label index.
        """
        label_id = len(self._evaluator.index_to_label)
        if label not in self._evaluator.label_to_index:
            self._evaluator.label_to_index[label] = label_id
            self._evaluator.index_to_label[label_id] = label

            label_id += 1

        return self._evaluator.label_to_index[label]

    def _add_data(
        self,
        uid_index: int,
        groundtruth: int,
        predictions: list[tuple[int, float]],
    ):

        pairs = list()
        scores = np.array([score for _, score in predictions])
        max_score_idx = np.argmax(scores)

        for idx, (plabel, score) in enumerate(predictions):
            pairs.append(
                (
                    float(uid_index),
                    float(groundtruth),
                    float(plabel),
                    float(score),
                    float(max_score_idx == idx),
                )
            )

        if self._evaluator._detailed_pairs.size == 0:
            self._evaluator._detailed_pairs = np.array(pairs)
        else:
            self._evaluator._detailed_pairs = np.concatenate(
                [
                    self._evaluator._detailed_pairs,
                    np.array(pairs),
                ],
                axis=0,
            )

    def add_data(
        self,
        classifications: list[Classification],
        show_progress: bool = False,
    ):
        """
        Adds classifications to the cache.

        Parameters
        ----------
        classifications : list[Classification]
            A list of Classification objects.
        show_progress : bool, default=False
            Toggle for tqdm progress bar.
        """

        disable_tqdm = not show_progress
        for classification in tqdm(classifications, disable=disable_tqdm):

            if len(classification.predictions) == 0:
                raise ValueError(
                    "Classifications must contain at least one prediction."
                )
            # update metadata
            self._evaluator.n_datums += 1
            self._evaluator.n_groundtruths += 1
            self._evaluator.n_predictions += len(classification.predictions)

            # update datum uid index
            uid_index = self._add_datum(uid=classification.uid)

            # cache labels and annotations
            groundtruth = self._add_label(classification.groundtruth)
            self.groundtruth_count[groundtruth][uid_index] += 1

            predictions = list()
            for plabel, pscore in zip(
                classification.predictions, classification.scores
            ):
                label_idx = self._add_label(plabel)
                self.prediction_count[label_idx][uid_index] += 1
                predictions.append(
                    (
                        label_idx,
                        pscore,
                    )
                )

            self._add_data(
                uid_index=uid_index,
                groundtruth=groundtruth,
                predictions=predictions,
            )

    def finalize(self) -> Evaluator:
        """
        Performs data finalization and some preprocessing steps.

        Returns
        -------
        Evaluator
            A ready-to-use evaluator object.
        """

        if self._evaluator._detailed_pairs.size == 0:
            raise ValueError("No data available to create evaluator.")

        n_datums = self._evaluator.n_datums
        n_labels = len(self._evaluator.index_to_label)

        self._evaluator.n_labels = n_labels

        self._evaluator._label_metadata_per_datum = np.zeros(
            (2, n_datums, n_labels), dtype=np.int32
        )
        for datum_idx in range(n_datums):
            for label_idx in range(n_labels):
                gt_count = (
                    self.groundtruth_count[label_idx].get(datum_idx, 0)
                    if label_idx in self.groundtruth_count
                    else 0
                )
                pd_count = (
                    self.prediction_count[label_idx].get(datum_idx, 0)
                    if label_idx in self.prediction_count
                    else 0
                )
                self._evaluator._label_metadata_per_datum[
                    :, datum_idx, label_idx
                ] = np.array([gt_count, pd_count])

        self._evaluator._label_metadata = np.array(
            [
                [
                    np.sum(
                        self._evaluator._label_metadata_per_datum[
                            0, :, label_idx
                        ]
                    ),
                    np.sum(
                        self._evaluator._label_metadata_per_datum[
                            1, :, label_idx
                        ]
                    ),
                ]
                for label_idx in range(n_labels)
            ],
            dtype=np.int32,
        )

        # sort pairs by groundtruth, prediction, score
        indices = np.lexsort(
            (
                self._evaluator._detailed_pairs[:, 1],
                self._evaluator._detailed_pairs[:, 2],
                -self._evaluator._detailed_pairs[:, 3],
            )
        )
        self._evaluator._detailed_pairs = self._evaluator._detailed_pairs[
            indices
        ]

        return self._evaluator



import numpy as np
from valor_lite.semantic_segmentation import DataLoader, Segmentation


def test_filtering(segmentations_from_boxes: list[Segmentation]):

    loader = DataLoader()
    loader.add_data(segmentations_from_boxes)
    evaluator = loader.finalize()

    assert evaluator.metadata == {
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
        "number_of_datums": 2,
        "number_of_labels": 2,
        "number_of_groundtruths": 2,
        "number_of_predictions": 2,
        "number_of_groundtruth_pixels": 540000,
        "number_of_prediction_pixels": 540000,
    }

    assert evaluator.n_datums == 2
    assert (
        evaluator._label_metadata == np.array([[10000, 10000], [15000, 5000]])
    ).all()

    # test datum filtering
    filter_by_uid1 = evaluator.create_filter(datum_uids=["uid1"])
    assert filter_by_uid1.indices == np.array([0])
    assert (
        filter_by_uid1.label_metadata == np.array([[10000, 10000], [0, 0]])
    ).all()

    filter_by_uid2 = evaluator.create_filter(datum_uids=["uid2"])
    assert filter_by_uid2.indices == np.array([1])
    assert (
        filter_by_uid2.label_metadata == np.array([[0, 0], [15000, 5000]])
    ).all()

    # test label filtering
    filter_by_label_v1 = evaluator.create_filter(labels=["v1"])
    assert (filter_by_label_v1.indices == np.array([0, 1])).all()
    assert (
        filter_by_label_v1.label_metadata == np.array([[10000, 10000], [0, 0]])
    ).all()

    filter_by_label_v2 = evaluator.create_filter(labels=["v2"])
    assert (filter_by_label_v2.indices == np.array([0, 1])).all()
    assert (
        filter_by_label_v2.label_metadata == np.array([[0, 0], [15000, 5000]])
    ).all()

    # test joint filtering
    filter_by_uid0_v1 = evaluator.create_filter(
        datum_uids=["uid1"], labels=["v1"]
    )
    assert filter_by_uid0_v1.indices == np.array([0])
    assert (
        filter_by_uid0_v1.label_metadata == np.array([[10000, 10000], [0, 0]])
    ).all()

    filter_by_uid0_v2 = evaluator.create_filter(
        datum_uids=["uid1"], labels=["v2"]
    )
    assert filter_by_uid0_v2.indices == np.array([0])
    assert (
        filter_by_uid0_v2.label_metadata == np.array([[0, 0], [0, 0]])
    ).all()

    # test filter all
    filter_by_all = evaluator.create_filter(datum_uids=[])
    assert filter_by_all.indices.size == 0
    assert (
        filter_by_uid0_v2.label_metadata == np.array([[0, 0], [0, 0]])
    ).all()



import pytest
from valor_lite.classification import Classification, DataLoader


def test_no_data():
    loader = DataLoader()
    with pytest.raises(ValueError):
        loader.finalize()


def test_missing_predictions(
    classifications_no_predictions: list[Classification],
):
    loader = DataLoader()

    with pytest.raises(ValueError) as e:
        loader.add_data(classifications_no_predictions)
    assert "Classifications must contain at least one prediction" in str(e)



import json
import os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from time import time

import requests
from tqdm import tqdm
from valor_lite.classification import Classification, DataLoader


def _convert_valor_dicts_into_Classification(gt_dict: dict, pd_dict: dict):
    """Convert a groundtruth dictionary and prediction dictionary into a valor_lite Classification object."""
    pds = []
    scores = []

    # there's only one annotation / label per groundtruth in the benchmarking data
    gt = gt_dict["annotations"][0]["labels"][0]["value"]
    pds = []
    scores = []
    for pann in pd_dict["annotations"]:
        for valor_label in pann["labels"]:
            pds.append(valor_label["value"])
            scores.append(valor_label["score"])

    return Classification(
        uid=gt_dict["datum"]["uid"],
        groundtruth=gt,
        predictions=pds,
        scores=scores,
    )


def time_it(fn):
    def wrapper(*args, **kwargs):
        start = time()
        results = fn(*args, **kwargs)
        return (time() - start, results)

    return wrapper


def download_data_if_not_exists(
    file_name: str,
    file_path: Path,
    url: str,
):
    """Download the data from a public bucket if it doesn't exist locally."""

    if not os.path.exists(file_path):
        response = requests.get(url, stream=True)
        if response.status_code == 200:
            total_size = int(response.headers.get("content-length", 0))
            with open(file_path, "wb") as f:
                with tqdm(
                    total=total_size,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                    desc=file_name,
                ) as pbar:
                    for chunk in response.iter_content(chunk_size=1024):
                        if chunk:
                            f.write(chunk)
                            pbar.update(1024)
        else:
            raise RuntimeError(response)
    else:
        print(f"{file_name} already exists locally.")


def write_results_to_file(write_path: Path, results: list[dict]):
    """Write results to results.json"""
    current_datetime = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
    if os.path.isfile(write_path):
        with open(write_path, "r") as file:
            file.seek(0)
            data = json.load(file)
    else:
        data = {}

    data[current_datetime] = results

    with open(write_path, "w+") as file:
        json.dump(data, file, indent=4)


@time_it
def ingest(
    loader: DataLoader,
    gt_path: Path,
    pd_path: Path,
    limit: int,
    chunk_size: int,
):
    accumulated_time = 0.0
    with open(gt_path, "r") as gf:
        with open(pd_path, "r") as pf:
            count = 0
            classifications = []
            for gline, pline in zip(gf, pf):

                gt_dict = json.loads(gline)
                pd_dict = json.loads(pline)
                classifications.append(
                    _convert_valor_dicts_into_Classification(
                        gt_dict=gt_dict, pd_dict=pd_dict
                    )
                )
                count += 1
                if count >= limit and limit > 0:
                    break
                elif len(classifications) < chunk_size or chunk_size == -1:
                    continue

                timer, _ = time_it(loader.add_data)(classifications)
                accumulated_time += timer
                classifications = []

            if classifications:
                timer, _ = time_it(loader.add_data)(classifications)
                accumulated_time += timer

    return accumulated_time


@dataclass
class Benchmark:
    limit: int
    n_datums: int
    n_groundtruths: int
    n_predictions: int
    n_labels: int
    chunk_size: int
    ingestion: float
    preprocessing: float
    precomputation: float
    evaluation: float
    detailed_evaluation: list[tuple[int, float]]

    def result(self) -> dict:
        return {
            "limit": self.limit,
            "n_datums": self.n_datums,
            "n_groundtruths": self.n_groundtruths,
            "n_predictions": self.n_predictions,
            "n_labels": self.n_labels,
            "chunk_size": self.chunk_size,
            "ingestion": {
                "loading_from_file": f"{round(self.ingestion - self.preprocessing, 2)} seconds",
                "numpy_conversion": f"{round(self.preprocessing, 2)} seconds",
                "finalization": f"{round(self.precomputation, 2)} seconds",
                "total": f"{round(self.ingestion + self.precomputation, 2)} seconds",
            },
            "base_evaluation": f"{round(self.evaluation, 2)} seconds",
            "detailed_evaluation": [
                {
                    "n_points": 10,
                    "n_examples": curve[0],
                    "computation": f"{round(curve[1], 2)} seconds",
                }
                for curve in self.detailed_evaluation
            ],
        }


def run_benchmarking_analysis(
    limits_to_test: list[int],
    results_file: str = "clf_results.json",
    chunk_size: int = -1,
    ingestion_timeout=30,
    evaluation_timeout=30,
):
    """Time various function calls and export the results."""
    current_directory = Path(__file__).parent
    write_path = current_directory / Path(results_file)

    gt_filename = "gt_classification.jsonl"
    pd_filename = "pd_classification.jsonl"

    # cache data locally
    for filename in [gt_filename, pd_filename]:
        file_path = current_directory / Path(filename)
        url = f"https://pub-fae71003f78140bdaedf32a7c8d331d2.r2.dev/{filename}"
        download_data_if_not_exists(
            file_name=filename, file_path=file_path, url=url
        )

    # iterate through datum limits
    results = list()
    for limit in limits_to_test:

        # === Base Evaluation ===
        loader = DataLoader()

        # ingest + preprocess
        (ingest_time, preprocessing_time,) = ingest(
            loader=loader,
            gt_path=current_directory / Path(gt_filename),
            pd_path=current_directory / Path(pd_filename),
            limit=limit,
            chunk_size=chunk_size,
        )  # type: ignore - time_it wrapper

        finalization_time, evaluator = time_it(loader.finalize)()

        if ingest_time > ingestion_timeout and ingestion_timeout != -1:
            raise TimeoutError(
                f"Base precomputation timed out with limit of {limit}."
            )

        # evaluate
        eval_time, _ = time_it(evaluator.compute_precision_recall_rocauc)()
        if eval_time > evaluation_timeout and evaluation_timeout != -1:
            raise TimeoutError(
                f"Base evaluation timed out with {evaluator.n_datums} datums."
            )

        detail_no_examples_time, _ = time_it(
            evaluator.compute_confusion_matrix
        )(
            number_of_examples=0,
        )
        if (
            detail_no_examples_time > evaluation_timeout
            and evaluation_timeout != -1
        ):
            raise TimeoutError(
                f"Base evaluation timed out with {evaluator.n_datums} datums."
            )

        detail_three_examples_time, _ = time_it(
            evaluator.compute_confusion_matrix
        )(
            number_of_examples=3,
        )
        if (
            detail_three_examples_time > evaluation_timeout
            and evaluation_timeout != -1
        ):
            raise TimeoutError(
                f"Base evaluation timed out with {evaluator.n_datums} datums."
            )

        results.append(
            Benchmark(
                limit=limit,
                n_datums=evaluator.n_datums,
                n_groundtruths=evaluator.n_groundtruths,
                n_predictions=evaluator.n_predictions,
                n_labels=evaluator.n_labels,
                chunk_size=chunk_size,
                ingestion=ingest_time,
                preprocessing=preprocessing_time,
                precomputation=finalization_time,
                evaluation=eval_time,
                detailed_evaluation=[
                    (0, detail_no_examples_time),
                    (3, detail_three_examples_time),
                ],
            ).result()
        )

    write_results_to_file(write_path=write_path, results=results)


if __name__ == "__main__":

    run_benchmarking_analysis(
        limits_to_test=[5000, 5000, 5000],
    )



import pytest
from valor_lite.text_generation import Context, Evaluator, QueryResponse
from valor_lite.text_generation.computation import calculate_context_recall


def test_context_recall_no_context(mock_client):

    evaluator = Evaluator(client=mock_client)
    with pytest.raises(ValueError):
        assert evaluator.compute_context_recall(
            response=QueryResponse(
                query="a",
                response="a",
            )
        )


def test_calculate_context_recall(
    mock_client,
    verdicts_all_yes: str,
    verdicts_all_no: str,
    verdicts_two_yes_one_no: str,
    verdicts_empty: str,
):

    evaluator = Evaluator(client=mock_client)

    mock_client.returning = verdicts_all_yes
    assert (
        calculate_context_recall(
            client=mock_client,
            system_prompt="",
            predicted_context=["a", "b", "c"],
            groundtruth_context=["x", "y", "z"],
        )
        == 1.0
    )
    assert evaluator.compute_context_recall(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=["x", "y", "z"],
                prediction=["a", "b", "c"],
            ),
        )
    ).to_dict() == {
        "type": "ContextRecall",
        "value": 1.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_two_yes_one_no
    assert (
        calculate_context_recall(
            client=mock_client,
            system_prompt="",
            predicted_context=["a", "b", "c"],
            groundtruth_context=["x", "y", "z"],
        )
        == 2 / 3
    )
    assert evaluator.compute_context_recall(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=["x", "y", "z"],
                prediction=["a", "b", "c"],
            ),
        )
    ).to_dict() == {
        "type": "ContextRecall",
        "value": 2 / 3,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_all_no
    assert (
        calculate_context_recall(
            client=mock_client,
            system_prompt="",
            predicted_context=["a", "b", "c"],
            groundtruth_context=["x", "y", "z"],
        )
        == 0.0
    )
    assert evaluator.compute_context_recall(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=["x", "y", "z"],
                prediction=["a", "b", "c"],
            ),
        )
    ).to_dict() == {
        "type": "ContextRecall",
        "value": 0.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_empty
    assert (
        calculate_context_recall(
            client=mock_client,
            system_prompt="",
            predicted_context=[],
            groundtruth_context=[],
        )
        == 1.0
    )
    assert evaluator.compute_context_recall(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=[],
                prediction=[],
            ),
        )
    ).to_dict() == {
        "type": "ContextRecall",
        "value": 1.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_empty
    assert (
        calculate_context_recall(
            client=mock_client,
            system_prompt="",
            predicted_context=["a"],
            groundtruth_context=[],
        )
        == 0.0
    )
    assert evaluator.compute_context_recall(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=[],
                prediction=["a"],
            ),
        )
    ).to_dict() == {
        "type": "ContextRecall",
        "value": 0.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_empty
    assert (
        calculate_context_recall(
            client=mock_client,
            system_prompt="",
            predicted_context=[],
            groundtruth_context=["b"],
        )
        == 0.0
    )
    assert evaluator.compute_context_recall(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(
                groundtruth=["x"],
                prediction=[],
            ),
        )
    ).to_dict() == {
        "type": "ContextRecall",
        "value": 0.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }



import pytest
from valor_lite.text_generation import Context, Evaluator, QueryResponse

try:
    import mistralai
except ImportError:
    mistralai = None

try:
    import openai
except ImportError:
    openai = None


@pytest.mark.skipif(
    openai is None,
    reason="Openai is not installed.",
)
def test_openai_integration():

    assert Evaluator.openai()

    with pytest.raises(ValueError) as e:
        Evaluator.openai(
            retries=1,
            seed=1,
        )
    assert "Seed is provided, but retries is not 0." in str(e)


@pytest.mark.skipif(
    mistralai is None,
    reason="MistralAI is not installed.",
)
def test_mistral_integration():
    assert Evaluator.mistral()


def test_compute_all(
    mock_client,
    verdicts_two_yes_one_no,
):
    mock_client.returning = verdicts_two_yes_one_no
    evaluator = Evaluator(client=mock_client)
    metrics = evaluator.compute_all(
        response=QueryResponse(
            query="The dog is wagging its tail.",
            response="The dog doesn't like the cat.",
            context=Context(
                groundtruth=[
                    "The dog has never met a cat.",
                    "The dog is happy to see the cat.",
                ],
                prediction=[
                    "The dog has never met a cat.",
                    "The dog is wagging its tail.",
                    "Cats and dogs are common pets.",
                ],
            ),
        )
    )

    actual = {
        mtype: [m.to_dict() for m in mvalues]
        for mtype, mvalues in metrics.items()
    }
    assert actual == {
        "AnswerCorrectness": [
            {
                "type": "AnswerCorrectness",
                "value": 0.8,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "AnswerRelevance": [
            {
                "type": "AnswerRelevance",
                "value": 0.6666666666666666,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "Bias": [
            {
                "type": "Bias",
                "value": 0.6666666666666666,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "ContextPrecision": [
            {
                "type": "ContextPrecision",
                "value": 0.8333333333333333,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "ContextRecall": [
            {
                "type": "ContextRecall",
                "value": 0.6666666666666666,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "ContextRelevance": [
            {
                "type": "ContextRelevance",
                "value": 0.6666666666666666,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "Faithfulness": [
            {
                "type": "Faithfulness",
                "value": 0.6666666666666666,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "Hallucination": [
            {
                "type": "Hallucination",
                "value": 0.6666666666666666,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "SummaryCoherence": [
            {
                "type": "SummaryCoherence",
                "value": 4,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "Toxicity": [
            {
                "type": "Toxicity",
                "value": 0.6666666666666666,
                "parameters": {"evaluator": "mock", "retries": 0},
            }
        ],
        "ROUGE": [
            {
                "type": "ROUGE",
                "value": 0.5333333333333333,
                "parameters": {"rouge_type": "rouge1", "use_stemmer": False},
            },
            {
                "type": "ROUGE",
                "value": 0.30769230769230765,
                "parameters": {"rouge_type": "rouge2", "use_stemmer": False},
            },
            {
                "type": "ROUGE",
                "value": 0.5333333333333333,
                "parameters": {"rouge_type": "rougeL", "use_stemmer": False},
            },
            {
                "type": "ROUGE",
                "value": 0.5333333333333333,
                "parameters": {
                    "rouge_type": "rougeLsum",
                    "use_stemmer": False,
                },
            },
        ],
        "BLEU": [
            {
                "type": "BLEU",
                "value": 0.0,
                "parameters": {"weights": [0.25, 0.25, 0.25, 0.25]},
            }
        ],
    }



from dataclasses import replace
from random import choice, uniform

import numpy as np
import pytest
from valor_lite.classification import Classification, DataLoader, MetricType


@pytest.fixture
def one_classification(
    basic_classifications: list[Classification],
) -> list[Classification]:
    assert len(basic_classifications) == 3
    return [basic_classifications[0]]


@pytest.fixture
def three_classifications(
    basic_classifications: list[Classification],
) -> list[Classification]:
    assert len(basic_classifications) == 3
    return basic_classifications


@pytest.fixture
def six_classifications(
    basic_classifications: list[Classification],
) -> list[Classification]:
    assert len(basic_classifications) == 3
    clf1 = basic_classifications[0]
    clf2 = basic_classifications[1]
    clf3 = basic_classifications[2]

    clf4 = replace(basic_classifications[0])
    clf5 = replace(basic_classifications[1])
    clf6 = replace(basic_classifications[2])

    clf4.uid = "uid4"
    clf5.uid = "uid5"
    clf6.uid = "uid6"

    return [clf1, clf2, clf3, clf4, clf5, clf6]


def generate_random_classifications(
    n_classifications: int, n_categories: int, n_labels: int
) -> list[Classification]:

    labels = [str(value) for value in range(n_labels)]

    return [
        Classification(
            uid=f"uid{i}",
            groundtruth=choice(labels),
            predictions=labels,
            scores=[uniform(0, 1) for _ in range(n_labels)],
        )
        for i in range(n_classifications)
    ]


def test_filtering_one_classification(
    one_classification: list[Classification],
):

    manager = DataLoader()
    manager.add_data(one_classification)
    evaluator = manager.finalize()

    assert evaluator._detailed_pairs.shape == (4, 5)
    assert (
        evaluator._detailed_pairs
        == np.array(
            [
                [0.0, 0.0, 0.0, 1.0, 1.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 3.0, 0.0, 0.0],
            ]
        )
    ).all()

    assert evaluator._label_metadata_per_datum.shape == (2, 1, 4)
    assert (
        evaluator._label_metadata_per_datum
        == np.array(
            [
                [
                    [1, 0, 0, 0],
                ],
                [[1, 1, 1, 1]],
            ]
        )
    ).all()

    assert (
        evaluator._label_metadata
        == np.array(
            [
                [
                    1,
                    1,
                ],
                [
                    0,
                    1,
                ],
                [
                    0,
                    1,
                ],
                [
                    0,
                    1,
                ],
            ]
        )
    ).all()

    # test datum filtering

    filter_ = evaluator.create_filter(datum_uids=["uid0"])
    assert (filter_.indices == np.array([0, 1, 2, 3])).all()
    assert (
        filter_.label_metadata
        == np.array(
            [
                [
                    1,
                    1,
                ],
                [
                    0,
                    1,
                ],
                [
                    0,
                    1,
                ],
                [
                    0,
                    1,
                ],
            ]
        )
    ).all()

    # test label filtering
    filter_ = evaluator.create_filter(labels=["0"])
    assert (filter_.indices == np.array([0, 1, 2, 3])).all()

    filter_ = evaluator.create_filter(labels=["1"])
    assert (filter_.indices == np.array([])).all()

    # test combo
    filter_ = evaluator.create_filter(
        datum_uids=["uid0"],
        labels=["0"],
    )
    assert (filter_.indices == np.array([0, 1, 2, 3])).all()

    # test evaluation
    filter_ = evaluator.create_filter(datum_uids=["uid0"])
    metrics = evaluator.evaluate(
        score_thresholds=[0.5],
        hardmax=False,
        filter_=filter_,
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "0",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "2",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "3",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_filtering_three_classifications(
    three_classifications: list[Classification],
):

    manager = DataLoader()
    manager.add_data(three_classifications)
    evaluator = manager.finalize()

    assert evaluator._detailed_pairs.shape == (12, 5)
    assert (
        evaluator._detailed_pairs
        == np.array(
            [
                [0.0, 0.0, 0.0, 1.0, 1.0],
                [1.0, 0.0, 2.0, 1.0, 1.0],
                [2.0, 3.0, 3.0, 0.3, 1.0],
                [1.0, 0.0, 0.0, 0.0, 0.0],
                [2.0, 3.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 0.0, 0.0],
                [2.0, 3.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 0.0, 0.0],
                [2.0, 3.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 3.0, 0.0, 0.0],
                [1.0, 0.0, 3.0, 0.0, 0.0],
            ]
        )
    ).all()

    assert evaluator._label_metadata_per_datum.shape == (2, 3, 4)
    assert (
        evaluator._label_metadata_per_datum
        == np.array(
            [
                [
                    [1, 0, 0, 0],
                    [1, 0, 0, 0],
                    [0, 0, 0, 1],
                ],
                [
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                ],
            ]
        )
    ).all()

    assert (
        evaluator._label_metadata == np.array([[2, 3], [0, 3], [0, 3], [1, 3]])
    ).all()

    # test datum filtering

    filter_ = evaluator.create_filter(datum_uids=["uid0"])
    assert (filter_.indices == np.array([0, 5, 8, 10])).all()
    assert (
        filter_.label_metadata == np.array([[1, 1], [0, 1], [0, 1], [0, 1]])
    ).all()

    filter_ = evaluator.create_filter(datum_uids=["uid2"])
    assert (filter_.indices == np.array([2, 4, 7, 9])).all()

    # test label filtering

    filter_ = evaluator.create_filter(labels=["0"])
    assert (filter_.indices == np.array([0, 1, 3, 5, 6, 8, 10, 11])).all()

    filter_ = evaluator.create_filter(labels=["1"])
    assert (filter_.indices == np.array([])).all()

    with pytest.raises(KeyError):
        filter_ = evaluator.create_filter(labels=["other"])

    # test combo
    filter_ = evaluator.create_filter(
        datum_uids=["uid0"],
        labels=["0"],
    )
    assert (filter_.indices == np.array([0, 5, 8, 10])).all()

    # test evaluation
    filter_ = evaluator.create_filter(datum_uids=["uid0"])
    metrics = evaluator.evaluate(
        score_thresholds=[0.5],
        hardmax=False,
        filter_=filter_,
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "0",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "2",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "3",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_filtering_six_classifications(
    six_classifications: list[Classification],
):

    manager = DataLoader()
    manager.add_data(six_classifications)
    evaluator = manager.finalize()

    assert evaluator._detailed_pairs.shape == (24, 5)
    assert (
        evaluator._detailed_pairs
        == np.array(
            [
                [0.0, 0.0, 0.0, 1.0, 1.0],
                [3.0, 0.0, 0.0, 1.0, 1.0],
                [1.0, 0.0, 2.0, 1.0, 1.0],
                [4.0, 0.0, 2.0, 1.0, 1.0],
                [2.0, 3.0, 3.0, 0.3, 1.0],
                [5.0, 3.0, 3.0, 0.3, 1.0],
                [1.0, 0.0, 0.0, 0.0, 0.0],
                [4.0, 0.0, 0.0, 0.0, 0.0],
                [2.0, 3.0, 0.0, 0.0, 0.0],
                [5.0, 3.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [1.0, 0.0, 1.0, 0.0, 0.0],
                [3.0, 0.0, 1.0, 0.0, 0.0],
                [4.0, 0.0, 1.0, 0.0, 0.0],
                [2.0, 3.0, 1.0, 0.0, 0.0],
                [5.0, 3.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 2.0, 0.0, 0.0],
                [3.0, 0.0, 2.0, 0.0, 0.0],
                [2.0, 3.0, 2.0, 0.0, 0.0],
                [5.0, 3.0, 2.0, 0.0, 0.0],
                [0.0, 0.0, 3.0, 0.0, 0.0],
                [1.0, 0.0, 3.0, 0.0, 0.0],
                [3.0, 0.0, 3.0, 0.0, 0.0],
                [4.0, 0.0, 3.0, 0.0, 0.0],
            ]
        )
    ).all()

    assert evaluator._label_metadata_per_datum.shape == (2, 6, 4)
    assert (
        evaluator._label_metadata_per_datum
        == np.array(
            [
                [
                    [1, 0, 0, 0],
                    [1, 0, 0, 0],
                    [0, 0, 0, 1],
                    [1, 0, 0, 0],
                    [1, 0, 0, 0],
                    [0, 0, 0, 1],
                ],
                [
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                    [1, 1, 1, 1],
                ],
            ]
        )
    ).all()

    assert (
        evaluator._label_metadata == np.array([[4, 6], [0, 6], [0, 6], [2, 6]])
    ).all()

    # test datum filtering

    filter_ = evaluator.create_filter(datum_uids=["uid0"])
    assert (filter_.indices == np.array([0, 10, 16, 20])).all()
    assert (
        filter_.label_metadata == np.array([[1, 1], [0, 1], [0, 1], [0, 1]])
    ).all()

    filter_ = evaluator.create_filter(datum_uids=["uid2"])
    assert (filter_.indices == np.array([4, 8, 14, 18])).all()

    # test label filtering

    filter_ = evaluator.create_filter(labels=["0"])
    assert (
        filter_.indices
        == np.array([0, 1, 2, 3, 6, 7, 10, 11, 12, 13, 16, 17, 20, 21, 22, 23])
    ).all()

    filter_ = evaluator.create_filter(labels=["1"])
    assert (filter_.indices == np.array([])).all()

    with pytest.raises(KeyError):
        evaluator.create_filter(labels=["other"])

    # test combo
    filter_ = evaluator.create_filter(
        datum_uids=["uid0"],
        labels=["0"],
    )
    assert (filter_.indices == np.array([0, 10, 16, 20])).all()

    # test evaluation
    filter_ = evaluator.create_filter(datum_uids=["uid0"])

    metrics = evaluator.evaluate(
        score_thresholds=[0.5],
        hardmax=False,
        filter_=filter_,
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Counts]]
    expected_metrics = [
        {
            "type": "Counts",
            "value": {
                "tp": 1,
                "fp": 0,
                "fn": 0,
                "tn": 0,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "0",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "1",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "2",
            },
        },
        {
            "type": "Counts",
            "value": {
                "tp": 0,
                "fp": 0,
                "fn": 0,
                "tn": 1,
            },
            "parameters": {
                "score_threshold": 0.5,
                "hardmax": False,
                "label": "3",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_filtering_random_classifications():
    loader = DataLoader()
    loader.add_data(generate_random_classifications(13, 2, 10))
    evaluator = loader.finalize()
    f = evaluator.create_filter(datum_uids=["uid0"])
    evaluator.evaluate(score_thresholds=[0.5], hardmax=False, filter_=f)



import numpy as np
from valor_lite.object_detection import (
    DataLoader,
    Detection,
    MetricType,
    compute_precion_recall,
)


def test_pr_curve_simple():

    sorted_pairs = np.array(
        [
            # dt,  gt,  pd,  iou,  gl,  pl, score,
            [0.0, 0.0, 2.0, 0.25, 0.0, 0.0, 0.95],
            [0.0, 0.0, 3.0, 0.33333, 0.0, 0.0, 0.9],
            [0.0, 0.0, 4.0, 0.66667, 0.0, 0.0, 0.65],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.1],
            [0.0, 0.0, 1.0, 0.5, 0.0, 0.0, 0.01],
        ]
    )

    label_metadata = np.array([[1, 5, 0]])
    iou_thresholds = np.array([0.1, 0.6])
    score_thresholds = np.array([0.0])

    (_, _, _, _, pr_curve) = compute_precion_recall(
        sorted_pairs,
        label_metadata=label_metadata,
        iou_thresholds=iou_thresholds,
        score_thresholds=score_thresholds,
    )

    assert pr_curve.shape == (2, 1, 101, 2)

    # test precision values
    assert np.isclose(pr_curve[0, 0, :, 0], 1.0).all()
    assert np.isclose(pr_curve[1, 0, :, 0], 1 / 3).all()

    # test score values
    assert np.isclose(pr_curve[0, 0, :, 1], 0.95).all()
    assert np.isclose(pr_curve[1, 0, :, 1], 0.65).all()


def test_pr_curve_using_torch_metrics_example(
    torchmetrics_detections: list[Detection],
):
    """
    cf with torch metrics/pycocotools results listed here:
    https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L231
    """
    loader = DataLoader()
    loader.add_bounding_boxes(torchmetrics_detections)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == ["3"]
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 4
    assert evaluator.n_labels == 6
    assert evaluator.n_groundtruths == 20
    assert evaluator.n_predictions == 19

    metrics = evaluator.evaluate(
        iou_thresholds=[0.5, 0.75],
    )

    # test PrecisionRecallCurve
    actual_metrics = [
        m.to_dict() for m in metrics[MetricType.PrecisionRecallCurve]
    ]
    expected_metrics = [
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": [1.0 for _ in range(101)],
                "scores": (
                    [0.953 for _ in range(21)]
                    + [0.805 for _ in range(20)]
                    + [0.611 for _ in range(20)]
                    + [0.407 for _ in range(20)]
                    + [0.335 for _ in range(20)]
                ),
            },
            "parameters": {
                "iou_threshold": 0.5,
                "label": "0",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": (
                    [1.0 for _ in range(41)]
                    + [0.8 for _ in range(40)]
                    + [0.0 for _ in range(20)]
                ),
                "scores": (
                    [0.953 for _ in range(21)]
                    + [0.805 for _ in range(20)]
                    + [0.407 for _ in range(20)]
                    + [0.335 for _ in range(20)]
                    + [0.0 for _ in range(20)]
                ),
            },
            "parameters": {
                "iou_threshold": 0.75,
                "label": "0",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": [1.0 for _ in range(101)],
                "scores": [0.3 for _ in range(101)],
            },
            "parameters": {
                "iou_threshold": 0.5,
                "label": "1",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": [1.0 for _ in range(101)],
                "scores": [0.3 for _ in range(101)],
            },
            "parameters": {
                "iou_threshold": 0.75,
                "label": "1",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": [1.0 for _ in range(51)]
                + [0.0 for _ in range(50)],
                "scores": [0.726 for _ in range(51)]
                + [0.0 for _ in range(50)],
            },
            "parameters": {
                "iou_threshold": 0.5,
                "label": "2",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": [1.0 for _ in range(51)]
                + [0.0 for _ in range(50)],
                "scores": [0.726 for _ in range(51)]
                + [0.0 for _ in range(50)],
            },
            "parameters": {
                "iou_threshold": 0.75,
                "label": "2",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": [1.0 for _ in range(101)],
                "scores": [0.546 for _ in range(51)]
                + [0.236 for _ in range(50)],
            },
            "parameters": {
                "iou_threshold": 0.5,
                "label": "4",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": [1.0 for _ in range(101)],
                "scores": [0.546 for _ in range(51)]
                + [0.236 for _ in range(50)],
            },
            "parameters": {
                "iou_threshold": 0.75,
                "label": "4",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": (
                    [1.0 for _ in range(71)]
                    + [8 / 9 for _ in range(10)]
                    + [0.0 for _ in range(20)]
                ),
                "scores": (
                    [0.883 for _ in range(11)]
                    + [0.782 for _ in range(10)]
                    + [0.561 for _ in range(10)]
                    + [0.532 for _ in range(10)]
                    + [0.349 for _ in range(10)]
                    + [0.271 for _ in range(10)]
                    + [0.204 for _ in range(10)]
                    + [0.202 for _ in range(10)]
                    + [0.0 for _ in range(20)]
                ),
            },
            "parameters": {
                "iou_threshold": 0.5,
                "label": "49",
            },
        },
        {
            "type": "PrecisionRecallCurve",
            "value": {
                "precisions": (
                    [1.0 for _ in range(41)]
                    + [0.8571428571428571 for _ in range(20)]
                    + [0.0 for _ in range(40)]
                ),
                "scores": (
                    [0.883 for _ in range(11)]
                    + [0.782 for _ in range(10)]
                    + [0.561 for _ in range(10)]
                    + [0.532 for _ in range(10)]
                    + [0.271 for _ in range(10)]
                    + [0.204 for _ in range(10)]
                    + [0.0 for _ in range(40)]
                ),
            },
            "parameters": {
                "iou_threshold": 0.75,
                "label": "49",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics



import pytest
from valor_lite.text_generation.llm.exceptions import InvalidLLMResponseError
from valor_lite.text_generation.llm.validators import (
    validate_statements,
    validate_verdicts,
)


def test_validate_statements():
    response = {
        "k1": ["a", "b", "c"],
        "k2": ["a"],
        "k3": "a",
        "k4": {"verdict": "a", "analysis": "b"},
        "k5": {"verdict": "a", "analysis": "b", "other": "c"},
    }

    # test if key is in response
    with pytest.raises(InvalidLLMResponseError) as e:
        validate_statements(
            response=response,
            key="other_key",
        )
    assert "did not include key" in str(e)

    # test proper formatting
    validate_statements(response=response, key="k1")
    validate_statements(response=response, key="k2")
    with pytest.raises(InvalidLLMResponseError):
        validate_statements(response=response, key="k3")
    with pytest.raises(InvalidLLMResponseError):
        validate_statements(response=response, key="k4")
    with pytest.raises(InvalidLLMResponseError):
        validate_statements(response=response, key="k5")

    # test restricted value sets
    validate_statements(
        response=response, key="k1", allowed_values={"a", "b", "c"}
    )
    with pytest.raises(InvalidLLMResponseError):
        validate_statements(response=response, key="k1", allowed_values={"a"})
    validate_statements(
        response=response, key="k2", allowed_values={"a", "b", "c"}
    )
    validate_statements(response=response, key="k2", allowed_values={"a"})

    # test length enforcement
    validate_statements(response=response, key="k1", enforce_length=3)
    with pytest.raises(InvalidLLMResponseError) as e:
        validate_statements(response=response, key="k1", enforce_length=1)
    assert "does not match input size" in str(e)


def test_validate_verdicts():

    response = {
        "k1": [
            {"verdict": "a", "analysis": "abcdef"},
            {"verdict": "b", "analysis": "abcdef"},
            {"verdict": "c", "analysis": "abcdef"},
        ],
        "k2": [
            {"verdict": "a", "analysis": "abcdef"},
            {"verdict": "a", "analysis": "abcdef"},
        ],
        "k3": "a",
        "k4": {"verdict": "a", "analysis": "b"},
        "k5": [{"verdict": "a", "analysis": "b", "other": "c"}],
        "k6": [1, 2, 3],
    }

    # test if key is in response
    with pytest.raises(InvalidLLMResponseError) as e:
        validate_verdicts(
            response=response,
            key="other_key",
        )
    assert "did not include key" in str(e)

    # test proper formatting
    validate_verdicts(response=response, key="k1")
    validate_verdicts(response=response, key="k2")
    with pytest.raises(InvalidLLMResponseError):
        validate_verdicts(response=response, key="k3")
    with pytest.raises(InvalidLLMResponseError):
        validate_verdicts(response=response, key="k4")
    with pytest.raises(InvalidLLMResponseError):
        validate_verdicts(response=response, key="k5")
    with pytest.raises(InvalidLLMResponseError):
        validate_verdicts(response=response, key="k6")

    # test allowed value enforcement
    validate_verdicts(
        response=response, key="k1", allowed_values={"a", "b", "c"}
    )
    with pytest.raises(InvalidLLMResponseError):
        validate_verdicts(response=response, key="k1", allowed_values={"a"})
    validate_verdicts(
        response=response, key="k2", allowed_values={"a", "b", "c"}
    )
    validate_verdicts(response=response, key="k2", allowed_values={"a"})

    # test length enforcement
    validate_verdicts(response=response, key="k1", enforce_length=3)
    with pytest.raises(InvalidLLMResponseError) as e:
        validate_verdicts(response=response, key="k1", enforce_length=1)
    assert "does not match input size" in str(e)



import json

import pytest
from valor_lite.text_generation import Context, Evaluator, QueryResponse
from valor_lite.text_generation.computation import calculate_faithfulness


def test_faithfulness_no_context(mock_client):

    evaluator = Evaluator(client=mock_client)
    with pytest.raises(ValueError):
        assert evaluator.compute_faithfulness(
            response=QueryResponse(
                query="a",
                response="a",
            )
        )


def test_faithfulness_no_claims(mock_client):
    mock_client.returning = json.dumps(
        {
            "claims": [],
        }
    )

    assert (
        calculate_faithfulness(
            client=mock_client,
            system_prompt="",
            response="a",
            context=["x", "y", "z"],
        )
        == 1.0
    )

    evaluator = Evaluator(client=mock_client)
    assert evaluator.compute_faithfulness(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(prediction=["x", "y", "z"]),
        )
    ).to_dict() == {
        "type": "Faithfulness",
        "value": 1.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }


def test_calculate_faithfulness(
    mock_client,
    verdicts_all_yes: str,
    verdicts_all_no: str,
    verdicts_two_yes_one_no: str,
    verdicts_empty: str,
):

    evaluator = Evaluator(client=mock_client)

    mock_client.returning = verdicts_all_yes
    assert (
        calculate_faithfulness(
            client=mock_client,
            system_prompt="",
            response="a",
            context=["x", "y", "z"],
        )
        == 1.0
    )
    assert evaluator.compute_faithfulness(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(prediction=["x", "y", "z"]),
        )
    ).to_dict() == {
        "type": "Faithfulness",
        "value": 1.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_two_yes_one_no
    assert (
        calculate_faithfulness(
            client=mock_client,
            system_prompt="",
            response="a",
            context=["x", "y", "z"],
        )
        == 2 / 3
    )
    assert evaluator.compute_faithfulness(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(prediction=["x", "y", "z"]),
        )
    ).to_dict() == {
        "type": "Faithfulness",
        "value": 2 / 3,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_all_no
    assert (
        calculate_faithfulness(
            client=mock_client,
            system_prompt="",
            response="a",
            context=["x", "y", "z"],
        )
        == 0.0
    )
    assert evaluator.compute_faithfulness(
        response=QueryResponse(
            query="a",
            response="a",
            context=Context(prediction=["x", "y", "z"]),
        )
    ).to_dict() == {
        "type": "Faithfulness",
        "value": 0.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_empty
    assert (
        calculate_faithfulness(
            client=mock_client,
            system_prompt="",
            response="a",
            context=[],
        )
        == 0.0
    )
    with pytest.raises(ValueError):
        evaluator.compute_faithfulness(
            response=QueryResponse(
                query="a",
                response="a",
            )
        )



import numpy as np
from valor_lite.object_detection import (
    DataLoader,
    Detection,
    Metric,
    MetricType,
)


def test_metadata_using_torch_metrics_example(
    torchmetrics_detections: list[Detection],
):
    """
    cf with torch metrics/pycocotools results listed here:
    https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L231
    """
    loader = DataLoader()
    loader.add_bounding_boxes(torchmetrics_detections)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == ["3"]
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 4
    assert evaluator.n_labels == 6
    assert evaluator.n_groundtruths == 20
    assert evaluator.n_predictions == 19

    assert evaluator.metadata == {
        "ignored_prediction_labels": ["3"],
        "missing_prediction_labels": [],
        "n_datums": 4,
        "n_labels": 6,
        "n_groundtruths": 20,
        "n_predictions": 19,
    }


def test_no_groundtruths(detections_no_groundtruths):

    loader = DataLoader()
    loader.add_bounding_boxes(detections_no_groundtruths)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == ["v1"]
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 2
    assert evaluator.n_labels == 1
    assert evaluator.n_groundtruths == 0
    assert evaluator.n_predictions == 2

    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.5],
    )

    assert len(metrics[MetricType.AP]) == 0


def test_no_predictions(detections_no_predictions):

    loader = DataLoader()
    loader.add_bounding_boxes(detections_no_predictions)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == []
    assert evaluator.missing_prediction_labels == ["v1"]
    assert evaluator.n_datums == 2
    assert evaluator.n_labels == 1
    assert evaluator.n_groundtruths == 2
    assert evaluator.n_predictions == 0

    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.5],
    )

    assert len(metrics[MetricType.AP]) == 1

    # test AP
    actual_metrics = [m.to_dict() for m in metrics[MetricType.AP]]
    expected_metrics = [
        {
            "type": "AP",
            "value": 0.0,
            "parameters": {
                "iou_threshold": 0.5,
                "label": "v1",
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def _flatten_metrics(m) -> list:
    if isinstance(m, dict):
        keys = list(m.keys())
        values = [
            inner_value
            for value in m.values()
            for inner_value in _flatten_metrics(value)
        ]
        return keys + values
    elif isinstance(m, list):
        return [
            inner_value
            for value in m
            for inner_value in _flatten_metrics(value)
        ]
    elif isinstance(m, Metric):
        return _flatten_metrics(m.to_dict())
    else:
        return [m]


def test_output_types_dont_contain_numpy(basic_detections: list[Detection]):
    manager = DataLoader()
    manager.add_bounding_boxes(basic_detections)
    evaluator = manager.finalize()

    metrics = evaluator.evaluate(
        score_thresholds=[0.25, 0.75],
    )

    values = _flatten_metrics(metrics)
    for value in values:
        if isinstance(value, (np.generic, np.ndarray)):
            raise TypeError(f"Value `{value}` has type `{type(value)}`.")



from valor_lite.profiling import Benchmark, BenchmarkError
from valor_lite.semantic_segmentation.benchmark import (
    benchmark_add_data,
    benchmark_evaluate,
    benchmark_finalize,
)


def benchmark(
    bitmask_shape: tuple[int, int],
    number_of_unique_labels: int,
    number_of_images: int,
    *_,
    memory_limit: float = 4.0,
    time_limit: float = 10.0,
    repeat: int = 1,
    verbose: bool = False,
):
    """
    Runs a single benchmark.

    Parameters
    ----------
    bitmask_shape : tuple[int, int]
        The size (h, w) of the bitmask array.
    number_of_unique_labels : int
        The number of unique labels used in the synthetic example.
    number_of_images : int
        The number of distinct datums that are created.
    memory_limit : float
        The maximum amount of system memory allowed in gigabytes (GB).
    time_limit : float
        The maximum amount of time permitted before killing the benchmark.
    repeat : int
        The number of times to run a benchmark to produce an average runtime.
    verbose : bool, default=False
        Toggles terminal output of benchmark results.
    """

    b = Benchmark(
        time_limit=time_limit,
        memory_limit=int(memory_limit * (1024**3)),
        repeat=repeat,
        verbose=verbose,
    )

    _, failed, details = b.run(
        benchmark=benchmark_add_data,
        n_labels=[number_of_unique_labels],
        shape=[bitmask_shape],
    )
    if failed:
        raise BenchmarkError(
            benchmark=details["benchmark"],
            error_type=failed[0]["error"],
            error_message=failed[0]["msg"],
        )

    _, failed, details = b.run(
        benchmark=benchmark_finalize,
        n_datums=[number_of_images],
        n_labels=[number_of_unique_labels],
    )
    if failed:
        raise BenchmarkError(
            benchmark=details["benchmark"],
            error_type=failed[0]["error"],
            error_message=failed[0]["msg"],
        )

    _, failed, details = b.run(
        benchmark=benchmark_evaluate,
        n_datums=[number_of_images],
        n_labels=[number_of_unique_labels],
    )
    if failed:
        raise BenchmarkError(
            benchmark=details["benchmark"],
            error_type=failed[0]["error"],
            error_message=failed[0]["msg"],
        )


if __name__ == "__main__":

    benchmark(
        bitmask_shape=(4000, 4000),
        number_of_images=1000,
        number_of_unique_labels=10,
        memory_limit=4.0,
        time_limit=10.0,
        repeat=1,
        verbose=True,
    )



import numpy as np
from valor_lite.semantic_segmentation import DataLoader, Metric, Segmentation


def test_metadata_using_large_random_segmentations(
    large_random_segmentations: list[Segmentation],
):
    manager = DataLoader()
    manager.add_data(large_random_segmentations)
    evaluator = manager.finalize()

    assert evaluator.ignored_prediction_labels == []
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 3
    assert evaluator.n_labels == 9
    assert evaluator.n_groundtruths == 9
    assert evaluator.n_predictions == 9
    assert evaluator.n_groundtruth_pixels == 36000000
    assert evaluator.n_prediction_pixels == 36000000

    assert evaluator.metadata == {
        "ignored_prediction_labels": [],
        "missing_prediction_labels": [],
        "number_of_datums": 3,
        "number_of_labels": 9,
        "number_of_groundtruths": 9,
        "number_of_predictions": 9,
        "number_of_groundtruth_pixels": 36000000,
        "number_of_prediction_pixels": 36000000,
    }


def _flatten_metrics(m) -> list:
    if isinstance(m, dict):
        keys = list(m.keys())
        values = [
            inner_value
            for value in m.values()
            for inner_value in _flatten_metrics(value)
        ]
        return keys + values
    elif isinstance(m, list):
        return [
            inner_value
            for value in m
            for inner_value in _flatten_metrics(value)
        ]
    elif isinstance(m, Metric):
        return _flatten_metrics(m.to_dict())
    else:
        return [m]


def test_output_types_dont_contain_numpy(
    segmentations_from_boxes: list[Segmentation],
):
    manager = DataLoader()
    manager.add_data(segmentations_from_boxes)
    evaluator = manager.finalize()

    metrics = evaluator.evaluate()

    values = _flatten_metrics(metrics)
    for value in values:
        if isinstance(value, (np.generic, np.ndarray)):
            raise TypeError(value)



import numpy as np
from valor_lite.object_detection import DataLoader, Detection, MetricType
from valor_lite.object_detection.computation import compute_precion_recall


def test__compute_average_precision():

    sorted_pairs = np.array(
        [
            # dt,  gt,  pd,  iou,  gl,  pl, score,
            [0.0, 0.0, 2.0, 0.25, 0.0, 0.0, 0.95],
            [0.0, 0.0, 3.0, 0.33333, 0.0, 0.0, 0.9],
            [0.0, 0.0, 4.0, 0.66667, 0.0, 0.0, 0.65],
            [0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.1],
            [0.0, 0.0, 1.0, 0.5, 0.0, 0.0, 0.01],
        ]
    )

    label_metadata = np.array([[1, 5, 0]])
    iou_thresholds = np.array([0.1, 0.6])
    score_thresholds = np.array([0.0])

    (_, _, accuracy, _, _) = compute_precion_recall(
        sorted_pairs,
        label_metadata=label_metadata,
        iou_thresholds=iou_thresholds,
        score_thresholds=score_thresholds,
    )

    expected = np.array(
        [
            [0.2],  # iou = 0.1
            [0.2],  # iou = 0.6
        ]
    )
    assert (accuracy == expected).all()


def test_ap_using_torch_metrics_example(
    torchmetrics_detections: list[Detection],
):
    """
    cf with torch metrics/pycocotools results listed here:
    https://github.com/Lightning-AI/metrics/blob/107dbfd5fb158b7ae6d76281df44bd94c836bfce/tests/unittests/detection/test_map.py#L231
    """

    loader = DataLoader()
    loader.add_bounding_boxes(torchmetrics_detections)
    evaluator = loader.finalize()

    assert evaluator.ignored_prediction_labels == ["3"]
    assert evaluator.missing_prediction_labels == []
    assert evaluator.n_datums == 4
    assert evaluator.n_labels == 6
    assert evaluator.n_groundtruths == 20
    assert evaluator.n_predictions == 19

    metrics = evaluator.evaluate(
        iou_thresholds=[0.5, 0.75],
    )

    # test Accuracy
    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 9 / 19,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.5,
            },
        },
        {
            "type": "Accuracy",
            "value": 8 / 19,
            "parameters": {
                "iou_threshold": 0.75,
                "score_threshold": 0.5,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_metrics_first_class(
    basic_detections_first_class: list[Detection],
    basic_rotated_detections_first_class: list[Detection],
):
    """
    Basic object detection test.

    groundtruths
        datum uid1
            box 1 - label v1 - tp
            box 3 - label v2 - fn missing prediction
        datum uid2
            box 2 - label v1 - fn missing prediction

    predictions
        datum uid1
            box 1 - label v1 - score 0.3 - tp
        datum uid2
            box 2 - label v2 - score 0.98 - fp
    """
    for input_, method in [
        (basic_detections_first_class, DataLoader.add_bounding_boxes),
        (basic_rotated_detections_first_class, DataLoader.add_polygons),
    ]:
        loader = DataLoader()
        method(loader, input_)
        evaluator = loader.finalize()

        metrics = evaluator.evaluate(
            iou_thresholds=[0.1, 0.6],
            score_thresholds=[0.0, 0.5],
        )

        assert evaluator.ignored_prediction_labels == []
        assert evaluator.missing_prediction_labels == []
        assert evaluator.n_datums == 2
        assert evaluator.n_labels == 1
        assert evaluator.n_groundtruths == 2
        assert evaluator.n_predictions == 1

        # test Accuracy
        actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
        expected_metrics = [
            {
                "type": "Accuracy",
                "value": 1.0,
                "parameters": {
                    "iou_threshold": 0.1,
                    "score_threshold": 0.0,
                },
            },
            {
                "type": "Accuracy",
                "value": 1.0,
                "parameters": {
                    "iou_threshold": 0.6,
                    "score_threshold": 0.0,
                },
            },
            {
                "type": "Accuracy",
                "value": 0.0,
                "parameters": {
                    "iou_threshold": 0.1,
                    "score_threshold": 0.5,
                },
            },
            {
                "type": "Accuracy",
                "value": 0.0,
                "parameters": {
                    "iou_threshold": 0.6,
                    "score_threshold": 0.5,
                },
            },
        ]
        for m in actual_metrics:
            assert m in expected_metrics
        for m in expected_metrics:
            assert m in actual_metrics


def test_accuracy_metrics_second_class(
    basic_detections_second_class: list[Detection],
    basic_rotated_detections_second_class: list[Detection],
):
    """
    Basic object detection test.

    groundtruths
        datum uid1
            box 3 - label v2 - fn missing prediction
        datum uid2
           none
    predictions
        datum uid1
            none
        datum uid2
            box 2 - label v2 - score 0.98 - fp
    """
    for input_, method in [
        (basic_detections_second_class, DataLoader.add_bounding_boxes),
        (basic_rotated_detections_second_class, DataLoader.add_polygons),
    ]:
        loader = DataLoader()
        method(loader, input_)
        evaluator = loader.finalize()

        metrics = evaluator.evaluate(
            iou_thresholds=[0.1, 0.6],
            score_thresholds=[0.0, 0.5],
        )

        assert evaluator.ignored_prediction_labels == []
        assert evaluator.missing_prediction_labels == []
        assert evaluator.n_datums == 2
        assert evaluator.n_labels == 1
        assert evaluator.n_groundtruths == 1
        assert evaluator.n_predictions == 1

        # test Accuracy
        actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
        expected_metrics = [
            {
                "type": "Accuracy",
                "value": 0.0,
                "parameters": {
                    "iou_threshold": 0.1,
                    "score_threshold": 0.0,
                },
            },
            {
                "type": "Accuracy",
                "value": 0.0,
                "parameters": {
                    "iou_threshold": 0.6,
                    "score_threshold": 0.0,
                },
            },
            {
                "type": "Accuracy",
                "value": 0.0,
                "parameters": {
                    "iou_threshold": 0.1,
                    "score_threshold": 0.5,
                },
            },
            {
                "type": "Accuracy",
                "value": 0.0,
                "parameters": {
                    "iou_threshold": 0.6,
                    "score_threshold": 0.5,
                },
            },
        ]
        for m in actual_metrics:
            assert m in expected_metrics
        for m in expected_metrics:
            assert m in actual_metrics


def test_accuracy_false_negatives_single_datum_baseline(
    false_negatives_single_datum_baseline_detections: list[Detection],
):
    """This is the baseline for the below test. In this case there are two predictions and
    one groundtruth, but the highest confident prediction overlaps sufficiently with the groundtruth
    so there is not a penalty for the false negative so the Accuracy is 1
    """

    loader = DataLoader()
    loader.add_bounding_boxes(false_negatives_single_datum_baseline_detections)
    evaluator = loader.finalize()

    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.0, 0.9],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.0,
            },
        },
        {
            "type": "Accuracy",
            "value": 0.0,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.9,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_false_negatives_single_datum(
    false_negatives_single_datum_detections: list[Detection],
):
    """Tests where high confidence false negative was not being penalized. The
    difference between this test and the above is that here the prediction with higher confidence
    does not sufficiently overlap the groundtruth and so is penalized and we get an Accuracy of 0.5
    """

    loader = DataLoader()
    loader.add_bounding_boxes(false_negatives_single_datum_detections)
    evaluator = loader.finalize()
    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.0],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.0,
            },
        }
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_false_negatives_two_datums_one_empty_low_confidence_of_fp(
    false_negatives_two_datums_one_empty_low_confidence_of_fp_detections: list[
        Detection
    ],
):
    """In this test we have
        1. An image with a matching groundtruth and prediction (same class and high IOU)
        2. A second image with empty groundtruth annotation but a prediction with lower confidence
        then the prediction on the first image.

    In this case, the Accuracy should be 1.0 since the false positive has lower confidence than the true positive

    """

    loader = DataLoader()
    loader.add_bounding_boxes(
        false_negatives_two_datums_one_empty_low_confidence_of_fp_detections
    )
    evaluator = loader.finalize()
    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.0],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.0,
            },
        }
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_false_negatives_two_datums_one_empty_high_confidence_of_fp(
    false_negatives_two_datums_one_empty_high_confidence_of_fp_detections: list[
        Detection
    ],
):
    """In this test we have
        1. An image with a matching groundtruth and prediction (same class and high IOU)
        2. A second image with empty groundtruth annotation and a prediction with higher confidence
        then the prediction on the first image.

    In this case, the Accuracy should be 0.5 since the false positive has higher confidence than the true positive
    """

    loader = DataLoader()
    loader.add_bounding_boxes(
        false_negatives_two_datums_one_empty_high_confidence_of_fp_detections
    )
    evaluator = loader.finalize()
    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.0],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.0,
            },
        }
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_false_negatives_two_datums_one_only_with_different_class_low_confidence_of_fp(
    false_negatives_two_datums_one_only_with_different_class_low_confidence_of_fp_detections: list[
        Detection
    ],
):
    """In this test we have
        1. An image with a matching groundtruth and prediction (same class, `"value"`, and high IOU)
        2. A second image with a groundtruth annotation with class `"other value"` and a prediction with lower confidence
        then the prediction on the first image.

    In this case, the Accuracy for class `"value"` should be 1 since the false positive has lower confidence than the true positive.
    Accuracy for class `"other value"` should be 0 since there is no prediction for the `"other value"` groundtruth
    """
    loader = DataLoader()
    loader.add_bounding_boxes(
        false_negatives_two_datums_one_only_with_different_class_low_confidence_of_fp_detections
    )
    evaluator = loader.finalize()
    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.0],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.0,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics


def test_accuracy_false_negatives_two_datums_one_only_with_different_class_high_confidence_of_fp(
    false_negatives_two_images_one_only_with_different_class_high_confidence_of_fp_detections: list[
        Detection
    ],
):
    """In this test we have
        1. An image with a matching groundtruth and prediction (same class, `"value"`, and high IOU)
        2. A second image with a groundtruth annotation with class `"other value"` and a prediction with higher confidence
        then the prediction on the first image.

    In this case, the Accuracy for class `"value"` should be 0.5 since the false positive has higher confidence than the true positive.
    Accuracy for class `"other value"` should be 0 since there is no prediction for the `"other value"` groundtruth
    """
    loader = DataLoader()
    loader.add_bounding_boxes(
        false_negatives_two_images_one_only_with_different_class_high_confidence_of_fp_detections
    )
    evaluator = loader.finalize()
    metrics = evaluator.evaluate(
        iou_thresholds=[0.5],
        score_thresholds=[0.0],
    )

    actual_metrics = [m.to_dict() for m in metrics[MetricType.Accuracy]]
    expected_metrics = [
        {
            "type": "Accuracy",
            "value": 0.5,
            "parameters": {
                "iou_threshold": 0.5,
                "score_threshold": 0.0,
            },
        },
    ]
    for m in actual_metrics:
        assert m in expected_metrics
    for m in expected_metrics:
        assert m in actual_metrics



from valor_lite.text_generation import Evaluator, QueryResponse
from valor_lite.text_generation.computation import calculate_answer_relevance


def test_calculate_answer_relevance(
    mock_client,
    verdicts_all_yes: str,
    verdicts_all_no: str,
    verdicts_two_yes_one_no: str,
    verdicts_empty: str,
):

    evaluator = Evaluator(client=mock_client)

    mock_client.returning = verdicts_all_yes
    assert (
        calculate_answer_relevance(
            client=mock_client,
            system_prompt="",
            query="a",
            response="a",
        )
        == 1.0
    )
    assert evaluator.compute_answer_relevance(
        response=QueryResponse(
            query="a",
            response="a",
        )
    ).to_dict() == {
        "type": "AnswerRelevance",
        "value": 1.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_two_yes_one_no
    assert (
        calculate_answer_relevance(
            client=mock_client,
            system_prompt="",
            query="a",
            response="a",
        )
        == 2 / 3
    )
    assert evaluator.compute_answer_relevance(
        response=QueryResponse(
            query="a",
            response="a",
        )
    ).to_dict() == {
        "type": "AnswerRelevance",
        "value": 2 / 3,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_all_no
    assert (
        calculate_answer_relevance(
            client=mock_client,
            system_prompt="",
            query="a",
            response="a",
        )
        == 0.0
    )
    assert evaluator.compute_answer_relevance(
        response=QueryResponse(
            query="a",
            response="a",
        )
    ).to_dict() == {
        "type": "AnswerRelevance",
        "value": 0.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }

    mock_client.returning = verdicts_empty
    assert (
        calculate_answer_relevance(
            client=mock_client,
            system_prompt="",
            query="a",
            response="a",
        )
        == 0.0
    )
    assert evaluator.compute_answer_relevance(
        response=QueryResponse(
            query="a",
            response="a",
        )
    ).to_dict() == {
        "type": "AnswerRelevance",
        "value": 0.0,
        "parameters": {
            "evaluator": "mock",
            "retries": 0,
        },
    }



import numpy as np
import pytest
from valor_lite.semantic_segmentation import Bitmask, Segmentation


def _generate_boolean_mask(
    mask_shape: tuple[int, int],
    annotation_shape: tuple[int, int, int, int],
    label: str,
) -> Bitmask:
    mask = np.zeros(mask_shape, dtype=np.bool_)
    xmin = annotation_shape[0]
    xmax = annotation_shape[1]
    ymin = annotation_shape[2]
    ymax = annotation_shape[3]
    mask[ymin:ymax, xmin:xmax] = True
    return Bitmask(
        mask=mask,
        label=label,
    )


def _generate_random_boolean_mask(
    mask_shape: tuple[int, int],
    infill: float,
    label: str,
) -> Bitmask:
    mask_size = mask_shape[0] * mask_shape[1]
    mask = np.zeros(mask_size, dtype=np.bool_)

    n_positives = int(mask_size * infill)
    mask[:n_positives] = True
    np.random.shuffle(mask)
    mask = mask.reshape(mask_shape)

    return Bitmask(
        mask=mask,
        label=label,
    )


@pytest.fixture
def basic_segmentations() -> list[Segmentation]:

    gmask1 = Bitmask(
        mask=np.array([[True, False], [False, True]]),
        label="v1",
    )
    gmask2 = Bitmask(
        mask=np.array([[False, False], [True, False]]),
        label="v2",
    )

    pmask1 = Bitmask(
        mask=np.array([[True, False], [False, False]]),
        label="v1",
    )
    pmask2 = Bitmask(
        mask=np.array([[False, True], [True, False]]),
        label="v2",
    )

    return [
        Segmentation(
            uid="uid0",
            groundtruths=[gmask1, gmask2],
            predictions=[pmask1, pmask2],
        )
    ]


@pytest.fixture
def segmentations_from_boxes() -> list[Segmentation]:

    mask_shape = (900, 300)

    rect1 = (0, 100, 0, 100)
    rect2 = (150, 300, 400, 500)
    rect3 = (50, 150, 0, 100)  # overlaps 50% with rect1
    rect4 = (101, 151, 301, 401)  # overlaps 1 pixel with rect2

    gmask1 = _generate_boolean_mask(mask_shape, rect1, "v1")
    gmask2 = _generate_boolean_mask(mask_shape, rect2, "v2")

    pmask1 = _generate_boolean_mask(mask_shape, rect3, "v1")
    pmask2 = _generate_boolean_mask(mask_shape, rect4, "v2")

    return [
        Segmentation(
            uid="uid1",
            groundtruths=[gmask1],
            predictions=[pmask1],
        ),
        Segmentation(
            uid="uid2",
            groundtruths=[gmask2],
            predictions=[pmask2],
        ),
    ]


@pytest.fixture
def large_random_segmentations() -> list[Segmentation]:

    mask_shape = (2000, 2000)
    infills_per_seg = [
        (0.9, 0.09, 0.01),
        (0.4, 0.4, 0.1),
        (0.3, 0.3, 0.3),
    ]
    labels_per_seg = [
        ("v1", "v2", "v3"),
        ("v4", "v5", "v6"),
        ("v7", "v8", "v9"),
    ]

    return [
        Segmentation(
            uid=f"uid{idx}",
            groundtruths=[
                _generate_random_boolean_mask(mask_shape, infill, label)
                for infill, label in zip(infills, labels)
            ],
            predictions=[
                _generate_random_boolean_mask(mask_shape, infill, label)
                for infill, label in zip(infills, labels)
            ],
        )
        for idx, (infills, labels) in enumerate(
            zip(infills_per_seg, labels_per_seg)
        )
    ]
